/*
 * Example 7: pBuffers
 * 
 * This example demonstrates the use of an OpenGL offscreen drawable
 * (pBuffer).  The pBuffer is used as a texture in the main drawable.
 * A teapot is drawn to the PBuffer and the pBuffer is shown as a 
 * texture on a cube.
 * 
 * Author: David Wolff
 * 
 * Licensed under the Creative Commons Attribution License 2.5:
 * http://creativecommons.org/licenses/by/2.5/
 */

package tutorial.g_PBuffers;

import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLCanvas;
import javax.media.opengl.GLCapabilities;
import javax.media.opengl.GLDrawableFactory;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.GLPbuffer;
import javax.media.opengl.glu.GLU;
import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JPanel;

import com.sun.opengl.util.FPSAnimator;

public class PBufferExample extends JFrame implements GLEventListener, ActionListener {
	
	// The animator object
	private FPSAnimator anim;
	// Frames per second
	private int fps = 30;
	// The rotation angle
	private double angle = 0.0;
	
	// Start and stop buttons
	private JButton startStopButton;
	
	// Texture ID
	private int pBufferTexID;
	
	// Vertex data for the cube
	private float[][] cubeData = { {-0.5f, -0.5f, 0.5f}, {0.5f, -0.5f, 0.5f},
			{0.5f, 0.5f, 0.5f}, {-0.5f, 0.5f, 0.5f}, {-0.5f, -0.5f, -0.5f}, 
			{0.5f, -0.5f, -0.5f}, {0.5f, 0.5f, -0.5f}, {-0.5f, 0.5f, -0.5f}
	};
	
	// The PBuffer object
	private GLPbuffer buffer;
	
	private GLU glu;
	
	/* 
	 * Create the pBuffer if the current OpenGL context allows it.
	 * Set up basic lighting, and turn on depth test.
	 */
	public void init( GLAutoDrawable d ) { 
		GL gl = d.getGL();
		gl.glClearColor( 0,0,0,0 );
		
		glu = new GLU();
		
		float[] lightAmbient = {0.4f, 0.4f, 0.4f, 1.0f};
		float[] lightDiffuse = {0.8f, 0.8f, 0.8f, 1.0f};
		float[] lightSpecular = {1.0f, 1.0f, 1.0f, 1.0f};
		float[] lightPosition = {0.0f, 0.0f, 10.0f, 1.0f}; 
		
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_AMBIENT, lightAmbient, 0);
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_DIFFUSE, lightDiffuse, 0);
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_SPECULAR, lightSpecular, 0);
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, lightPosition, 0);
		
		gl.glEnable(GL.GL_DEPTH_TEST);
		gl.glEnable(GL.GL_LIGHT0);
		gl.glEnable(GL.GL_LIGHTING);
		
		gl.glEnable(GL.GL_TEXTURE_2D);
		gl.glCullFace(GL.GL_BACK);
		gl.glEnable(GL.GL_CULL_FACE);
		gl.glMatrixMode( GL.GL_MODELVIEW );
		gl.glLoadIdentity();
		
		// Create the texture ID for the offscreen drawable
		int[] tmp = new int[1];
		gl.glGenTextures(1, tmp, 0);
		pBufferTexID = tmp[0];
		
		// Create the offscreen drawable (pBuffer).  Note that the width
		// and height must be a power of two if it is to be used as a 
		// texture.
		if( GLDrawableFactory.getFactory().canCreateGLPbuffer() ) {
			GLCapabilities caps = new GLCapabilities();
			buffer = GLDrawableFactory.getFactory().createGLPbuffer(
					caps,
					null,
					128,128,
					d.getContext()
				);
			buffer.addGLEventListener(new PBufferGLEventListener(pBufferTexID,128,128,glu));
		} else {
			System.err.println("Can't create a pBuffer.");
			System.exit(1);
		}
		
		// Create the Animator and start it
		anim = new FPSAnimator(d,fps);
		anim.start();
	}
	
	/* 
	 * Draw the scene.
	 */
	public void display( GLAutoDrawable d ) { 
		// Update the angle, and the time for the next frame
		angle += 1.0;
		if( angle > 180.0 ) angle -= 360.0;
		
		// Draw the frame
		GL gl = d.getGL();
		gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
		gl.glLoadIdentity();
		glu.gluLookAt(0,2,2,0,1.0,0,0,1,0);
		drawFloor(gl,true);
		gl.glTranslated(0,1,0);
		gl.glRotated(angle,0.5,0.5,-0.5);
		drawCube(gl);
	}
	
	/* 
	 * Set up a simple perspective projection, with appropriate aspect ratio.
	 */
	public void reshape( GLAutoDrawable d, int x, int y, int width, int height ) { 
		GL gl = d.getGL();
		gl.glMatrixMode( GL.GL_PROJECTION );
		gl.glLoadIdentity();
		glu.gluPerspective(60.0, (float)width/height, 0.5, 100.0);
		gl.glMatrixMode( GL.GL_MODELVIEW );
	}
	
	/*
	 * Draws a cube centered at the origin with side length 1.0.  The 
	 * pBuffer is used as the texture for all sides.  Note that we call
	 * display from the pBuffer here.  This is necessary to update the
	 * animation of the pBuffer.
	 */
	private void drawCube(GL gl)
	{
		// Update the pBuffer texture
		buffer.display();
		
		// Bind to the pBuffer texture
		gl.glBindTexture(GL.GL_TEXTURE_2D, pBufferTexID);
		
		// Draw the cube
		gl.glBegin(GL.GL_QUADS);
			// Front face
			gl.glNormal3f(0,0,1.0f);
			gl.glTexCoord2d(0,0);
			gl.glVertex3fv(cubeData[0],0);
			gl.glTexCoord2d(1,0);
			gl.glVertex3fv(cubeData[1],0);
			gl.glTexCoord2d(1,1);
			gl.glVertex3fv(cubeData[2],0);
			gl.glTexCoord2d(0,1);
			gl.glVertex3fv(cubeData[3],0);
			// Back face	
			gl.glNormal3f(0,0,-1.0f);
			gl.glTexCoord2d(0,0);
			gl.glVertex3fv(cubeData[5],0);
			gl.glTexCoord2d(1,0);
			gl.glVertex3fv(cubeData[4],0);
			gl.glTexCoord2d(1,1);
			gl.glVertex3fv(cubeData[7],0);
			gl.glTexCoord2d(0,1);
			gl.glVertex3fv(cubeData[6],0);
			// Top face
			gl.glNormal3f(0,1.0f,0.0f);
			gl.glTexCoord2d(0,0);
			gl.glVertex3fv(cubeData[3],0);
			gl.glTexCoord2d(1,0);
			gl.glVertex3fv(cubeData[2],0);
			gl.glTexCoord2d(1,1);
			gl.glVertex3fv(cubeData[6],0);
			gl.glTexCoord2d(0,1);
			gl.glVertex3fv(cubeData[7],0);
			// Left face
			gl.glNormal3f(-1.0f, 0.0f,0.0f);
			gl.glTexCoord2d(0,0);
			gl.glVertex3fv(cubeData[4],0);
			gl.glTexCoord2d(1,0);
			gl.glVertex3fv(cubeData[0],0);
			gl.glTexCoord2d(1,1);
			gl.glVertex3fv(cubeData[3],0);
			gl.glTexCoord2d(0,1);
			gl.glVertex3fv(cubeData[7],0);
			// Right face
			gl.glNormal3f(1.0f,0.0f,0.0f);
			gl.glTexCoord2d(0,0);
			gl.glVertex3fv(cubeData[1],0);
			gl.glTexCoord2d(1,0);
			gl.glVertex3fv(cubeData[5],0);
			gl.glTexCoord2d(1,1);
			gl.glVertex3fv(cubeData[6],0);
			gl.glTexCoord2d(0,1);
			gl.glVertex3fv(cubeData[2],0);
			// Bottom face
			gl.glNormal3f(0,-1.0f,0.0f);
			gl.glTexCoord2d(0,0);
			gl.glVertex3fv(cubeData[4],0);
			gl.glTexCoord2d(1,0);
			gl.glVertex3fv(cubeData[5],0);
			gl.glTexCoord2d(1,1);
			gl.glVertex3fv(cubeData[1],0);
			gl.glTexCoord2d(0,1);
			gl.glVertex3fv(cubeData[0],0);
		gl.glEnd();
	}
	
	/*
	 * Start or stop the animator.
	 */
	public void actionPerformed( ActionEvent e ) 
	{
		if( startStopButton.getText() == "Start") 
		{
			anim.start();
			startStopButton.setText("Stop");
		}
		else if( startStopButton.getText() == "Stop") {
			anim.stop();
			startStopButton.setText("Start");
		}
		
	}
	
	/*
	 * Create a JFrame with a GLCanvas and display it.
	 */
	public PBufferExample() 
	{
		setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );
		
		// Create a GLCanvas object and place into the JFrame
		GLCanvas canvas = new GLCanvas( );
		canvas.setSize(400,400);
		canvas.addGLEventListener( this );
		getContentPane().add( canvas, BorderLayout.CENTER );
		
		JPanel buttonPanel = new JPanel();
		buttonPanel.add( startStopButton = new JButton("Stop"));
		startStopButton.addActionListener(this);
		getContentPane().add( buttonPanel, BorderLayout.SOUTH );
		
		pack();
		setVisible( true );
	}
	
	/**
	 * Draws a tesselated 20 x 20 plane in the x-z plane, centered at the origin, 
	 * with 2500 quads.
	 * @param gl a GL object
	 * @param wireframe whether or not to draw in wireframe
	 */
	private void drawFloor(GL gl, boolean wireframe) {
		int xdivs = 50;
		int zdivs = 50;
		double xmin = -10.0, xmax = 10.0;
		double zmin = -10.0, zmax = 10.0;
		double x1, x2, y = 0.0, z;
		double xsize = (xmax - xmin) / xdivs;
		double zsize = (zmax - zmin) / zdivs;
		
		if( wireframe ) {
			gl.glPushAttrib(GL.GL_LIGHTING_BIT | GL.GL_POLYGON_BIT
					| GL.GL_COLOR_BUFFER_BIT | GL.GL_TEXTURE_BIT);
			gl.glDisable(GL.GL_LIGHTING);
			gl.glDisable(GL.GL_TEXTURE_2D);
			gl.glPolygonMode(GL.GL_FRONT_AND_BACK, GL.GL_LINE);
			gl.glColor3f(1.0f,0.2f,0.2f);
		}
		
		for( int i = 0; i < xdivs ; i++ ) {
			x1 = (i * xsize) + xmin;
			x2 = ((i+1) * xsize) + xmin;
			
			gl.glBegin(GL.GL_QUAD_STRIP);
			for(int j = 0; j <= zdivs; j++) {
				z = (j * zsize) + zmin;
				gl.glVertex3d(x2,y,z);
				gl.glVertex3d(x1,y,z);
			}
			gl.glEnd();
		}
		
		if(wireframe) {
			gl.glPopAttrib();
		}
	}
	
	public static void main( String[] args ) {
		new PBufferExample();
	}

	public void displayChanged( GLAutoDrawable d, boolean modeChanged, 
		boolean deviceChanged ) { }

}
