/*
 * Class PBufferGLEventListener
 *  
 * This is the GLEventListener for the pBuffer.  Draws a teapot, and
 * copies the pBuffer into the given texture.
 * 
 * Author: David Wolff
 * 
 * Licensed under the Creative Commons Attribution License 2.5:
 * http://creativecommons.org/licenses/by/2.5/
 */

package tutorial.g_PBuffers;

import javax.media.opengl.DebugGL;
import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;
import javax.media.opengl.GLEventListener;
import javax.media.opengl.glu.GLU;

import com.sun.opengl.util.GLUT;

public class PBufferGLEventListener implements GLEventListener {
	
	// The texture ID for this pBuffer
	private int texID;
	// Width and height of this pBuffer
	private int width, height;
	// A GLUT object for drawing the teapot
	private static GLUT glut = new GLUT();
	
	// Rotation angle
	private float angle;
	
	private GLU glu;
	
	/**
	 * Init the PBufferGLEventListener.
	 * @param id the texture ID for this pBuffer
	 * @param w the width of this pBuffer
	 * @param h the height of this pBuffer
	 */
	public PBufferGLEventListener( int id, int w, int h, GLU glu )
	{
		texID = id;
		width = w;
		height = h;
		angle = 0.0f;
		this.glu = glu;
	}
	
	/* 
	 * Set up lighting and initialize the texture by copying the current
	 * pBuffer (empty) into the texture.
	 * 
	 * Note:  we also set up the projection matrix here because the
	 * reshape method is not called in pBuffers.
	 */
	public void init(GLAutoDrawable d) {
		d.setGL( new DebugGL(d.getGL()));
		
		GL gl = d.getGL();
		gl.glClearColor( 0.5f, 0.5f, 0.8f, 1.0f );
		
		// Set up lighting
		float[] lightAmbient = {0.4f, 0.4f, 0.4f, 1.0f};
		float[] lightDiffuse = {0.8f, 0.8f, 0.8f, 1.0f};
		float[] lightSpecular = {1.0f, 1.0f, 1.0f, 1.0f};
		float[] lightPosition = {0.0f, 0.0f, 10.0f, 1.0f}; 
		
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_AMBIENT, lightAmbient,0);
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_DIFFUSE, lightDiffuse,0);
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_SPECULAR, lightSpecular,0);
		gl.glLightfv(GL.GL_LIGHT0, GL.GL_POSITION, lightPosition,0);
		
		gl.glEnable(GL.GL_DEPTH_TEST);
		gl.glEnable(GL.GL_LIGHT0);
		gl.glEnable(GL.GL_LIGHTING);
		
		// Set up the projection matrix.  Note, this is done here instead of
		// the reshape routine because in pBuffers, the reshape routine 
		// is never called.
		gl.glMatrixMode( GL.GL_PROJECTION );
		gl.glLoadIdentity();
		glu.gluPerspective( 60.0, (float)width/height, 0.4, 100 );
		gl.glMatrixMode( GL.GL_MODELVIEW );
		
		// Set up the texture parameters
		gl.glBindTexture(GL.GL_TEXTURE_2D, texID);
		
		gl.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_LINEAR);
		gl.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_LINEAR);
		gl.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_EDGE);
		gl.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_EDGE); 
		
		// Create the texture data by copying the current frame buffer.
		gl.glCopyTexImage2D(GL.GL_TEXTURE_2D,0,GL.GL_RGBA,0,0,width,height,0);
	}

	/* 
	 * Draw the teapot, and then copy the data from the buffer into the
	 * texture using glCopyTexSubImage2D.
	 */
	public void display(GLAutoDrawable d) 
	{	
		GL gl = d.getGL();
		
		gl.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT);
		
		gl.glMatrixMode(GL.GL_MODELVIEW);
		gl.glLoadIdentity();
		glu.gluLookAt(0,0,4,0,0,0,0,1,0);
		angle += 1.0f;
		if( angle > 180.0f ) angle -= 360.0f;
		gl.glRotatef(angle, 0,1,0);
		
		// Draw the pixels to the pBuffer
		glut.glutSolidTeapot(1.0);
		
		// Copy the pBuffer pixels into texture memory
		gl.glBindTexture(GL.GL_TEXTURE_2D, texID);
		gl.glCopyTexImage2D(GL.GL_TEXTURE_2D,0,GL.GL_RGBA,0,0,width,height,0);
		// gl.glCopyTexSubImage2D(GL.GL_TEXTURE_2D,0,0,0,0,0,width,height);
	}

	/* 
	 * Unused methods.
	 */
	public void reshape(GLAutoDrawable d, int x, int y, int width, int height) {}
	public void displayChanged(GLAutoDrawable arg0, boolean arg1, boolean arg2) {}

}
